/++
 + Firebird接続ライブラリ Version: 0.6.1
 +
 + 基本機能モジュール
 +
 + Copyright(C) 2009-2011 Takuya Kurosawa
 +/
module jp.ku6.firebird.core;

import std.string;
import std.datetime;
import org.firebirdsql.ibase;

/++
 + Firebird及びFirebirdAPIから発生したエラー
 +/
class RuntimeError : Error
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  status = FirebirdAPIでセットされるステータス
	 +  message = 追加メッセージ 省略可
	 +/
	this(ISC_STATUS_ARRAY status, string message = null)
	{
		char[] result;
		char[512] buffer;

		ISC_STATUS* statusPointer = status.ptr;

		if (message !is null)
			result = message.dup;

		for (;;)
		{
			int written = fb_interpret(buffer.ptr, buffer.length, &statusPointer);
			if (written == 0)
				break;
			else
				result ~= buffer[0 .. written];
		}

		super(cast(string)result);
	}

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  message = メッセージ
	 +/
	this(string message)
	{
		super(message);
	}
}

/++
 + データベース接続
 +/
class Connection
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  database = 接続先のデータベースの指定
	 +  user = 接続時に使用するユーザー名の指定
	 +  password = 接続時に使用するパスワードの指定
	 +/
	this(string database, string user, string password)
	in
	{
		assert(database.length > 0);
		assert(database.length < 0x8000);
		assert(user.length > 0);
		assert(user.length < 0x8000);
		assert(password.length < 0x8000);
	}
	body
	{
		// 接続パラメータを設定する
		if (isc_modify_dpb(&dpb_, &dpbLength_, isc_dpb_user_name, user.ptr, cast(short)user.length) != 0)
			throw new RuntimeError("Faild to modify DPB by user.");
		if (isc_modify_dpb(&dpb_, &dpbLength_, isc_dpb_password, password.ptr, cast(short)password.length) != 0)
			throw new RuntimeError("Faild to modify DPB by password.");

		database_ = database;
	}

	/++
	 + デストラクタ
	 +
	 + 接続が残っていた場合切断処理を行う
	 + とはいえ、GCで回収されるここの切断に頼るのはよろしくない
	 + あくまでどうしようも無いときの最終手段とすること
	 +/
	~this()
	{
		if(isAttached())
			detach();
	}
	
	/++
	 + 接続しているか確認する
	 +
	 + Returns:
	 +  true:接続中 false:未接続
	 +/
	const
	bool isAttached()
	{
		return (handle_ != 0);
	}

	/++
	 + 接続する
	 +
	 + Throws:
	 +  RuntimeError 接続に失敗した場合
	 +/
	void attach()
	in
	{
		assert(!isAttached());
	}
	body
	{
		if (isAttached())
			return;

		// 接続する
		ISC_STATUS_ARRAY status;
		isc_attach_database(status.ptr, cast(short)database_.length, database_.ptr, &handle_, dpbLength_, dpb_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);
	}

	/++
	 + 切断する
	 +
	 + Throws:
	 +  切断処理に失敗した場合
	 +/
	void detach()
	in
	{
		assert(isAttached());
	}
	body
	{
		// 接続済みか確認する
		if (!isAttached())
			return;

		// 接続を閉じる
		ISC_STATUS_ARRAY status;
		isc_detach_database(status.ptr, &handle_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);
		handle_ = 0;
	}

private:

	// 接続先データベース
	string database_;

	// Database Parameter Buffer
	ubyte* dpb_;

	// Database Parameter Bufferのサイズ
	short dpbLength_;

	// データベースハンドル
	isc_db_handle handle_;
}



/++
 + トランザクション
 +
 + Firebirdの仕様として、2相コミットも対応しているが、
 + 使うのはかなりのレアケースと判断し対応しない
 +/
class Transaction
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
	}

	/++
	 + コンストラクタ
	 +
	 + 使用するコネクションが決まっていれば予め指定出来る
	 +
	 + Params:
	 +  connection = 使用するコネクション
	 +/
	this(Connection connection)
	{
		connection_ = connection;
	}

	/++
	 + セットされているコネクションを取得
	 +
	 + Returns:
	 +  使用するコネクション
	 +/
	Connection connection()
	{
		return connection_;
	}

	/++
	 + 使用するコネクションを変更
	 +
	 + Params:
	 +  connection = 使用するコネクション
	 +/
	void connection(Connection newConnection)
	in
	{
		// トランザクション中は変更を許可しない
		assert(!begin_);
	}
	body
	{
		if (begin_)
			return;

		connection_ = newConnection;
	}

	/++
	 + トランザクションが開始されているか確認
	 +
	 + Returns:
	 +  true = 開始されている
	 +  false = 開始されていない
	 +/
	const
	bool isBegin()
	{
		return begin_;
	}

	/++
	 + 開始
	 +
	 + Throws:
	 +  RuntimeError 
	 +/
	void begin()
	in
	{
		assert(connection_ !is null);
	}
	body
	{
		if (connection_ is null)
			return;

		ISC_STATUS_ARRAY status;
		isc_start_transaction(status.ptr, &handle_, 1, &connection_.handle_, 0, 0);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		begin_ = true;
	}

	/++
	 + コミット
	 +
	 + Throws:
	 +  RuntimeError
	 +/
	void commit()
	in
	{
		assert(begin_);
	}
	body
	{
		if (!begin_)
			return;

		begin_ = false;

		ISC_STATUS_ARRAY status;
		isc_commit_transaction(status.ptr, &handle_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);
	}

	/++
	 + ロールバック
	 +
	 + Throws:
	 +  RuntimeError
	 +/
	void rollback()
	in
	{
		assert(begin_);
	}
	body
	{
		if (!begin_)
			return;

		begin_ = false;

		ISC_STATUS_ARRAY status;
		isc_rollback_transaction(status.ptr, &handle_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);
	}

private:

	// トランザクション開始フラグ
	bool begin_ = false;

	// トランザクションハンドル
	isc_tr_handle handle_;

	// 対象の接続オブジェクト
	Connection connection_;
}

/++
 + 問い合わせ結果
 +/
class ResultSet
{
	/++
	 + データベースから1レコード取得
	 +
	 + Returns:
	 +  true:	取得できた
	 +  false:	最後のレコードまで読んでいたので、取得できなかった
	 +
	 + Throws:
	 +  RuntimeError データベースからエラーが返された時
	 +/
	bool fetch()
	{
		ISC_STATUS_ARRAY status;
		if (isc_dsql_fetch(status.ptr, &statement_.handle_, 1, xsqlda_) == 100)
			return false;

		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		return true;
	}

	/++
	 + 指定カラムがNULLか確認する
	 +
	 + int等、D言語的にNULLを返せない場合にはこれで確認する
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  NULLの場合にtrueを返す
	 +/
	bool isNull(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;
		if ((var.sqltype & 1) == 0)
			return false;
		else
			return *(var.sqlind) == -1;
	}

	/++
	 + テキスト型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:string)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_TEXT || ((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_VARYING);
	}
	body
	{
		if (isNull(index))
			return null;

		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		short length;
		char* read;
		if ((var.sqltype & ~1) == SQL_TEXT)
		{
			length = var.sqllen;
			read = cast(char*)(var.sqldata);
		}
		else if ((var.sqltype & ~1) == SQL_VARYING)
		{
			length = *(cast(short*)var.sqldata);
			read = cast(char*)(var.sqldata) + 2;
		}
		else
		{
			assert(false);
		}

		char[] data = new char[length];
		char* write = data.ptr;

		for (short i; i<length; ++i)
			*(write++) = *(read++);

		return cast(invariant)data;
	}

	/++
	 + short型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:short)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_SHORT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(short*)var.sqldata);
	}

	/++
	 + ushort型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:ushort)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_SHORT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(ushort*)var.sqldata);
	}

	/++
	 + int型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:int)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_LONG);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(int*)var.sqldata);
	}

	/++
	 + uint型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:uint)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_LONG);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(uint*)var.sqldata);
	}
	
	/++
	 + long型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:long)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_INT64);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(long*)var.sqldata);
	}

	/++
	 + ulong型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:ulong)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_INT64);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(ulong*)var.sqldata);
	}

	/++
	 + float型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:float)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_FLOAT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(float*)var.sqldata);
	}

	/++
	 + double型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:double)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_DOUBLE);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		return *(cast(double*)var.sqldata);
	}

	/++
	 + timestamp型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:SysTime)(uint index)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) == 0 || *((xsqlda_.sqlvar.ptr + index).sqlind) == 0);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_TIMESTAMP);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		ISC_TIMESTAMP t = *(cast(ISC_TIMESTAMP*)var.sqldata);
		return SysTime((cast(long)(t.timestamp_date) + 678575L) * 24L * 60L * 60L * 10000000L + cast(long)(t.timestamp_time) * 1000L);
	}

	/++
	 + blob型のデータを取得
	 +
	 + Params:
	 +  index = 取得するカラムの位置
	 +  blob = 書き込み先のインスタンス指定(省略した場合はメソッド内でデフォルトコンストラクタを使用して生成)
	 +
	 + Returns:
	 +  カラム内のデータ
	 +/
	T get(T:Blob)(uint index, T blob = null)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_BLOB);
	}
	body
	{
		if (isNull(index))
			return null;

		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;
		ISC_QUAD* blobId = cast(ISC_QUAD*)var.sqldata;

		isc_blob_handle handle = 0;
		ISC_STATUS_ARRAY status;

		// blobをオープン
		isc_open_blob2(status.ptr, &statement_.transaction_.connection_.handle_, &statement_.transaction_.handle_, &handle, blobId, 0, null);
		if (status[1] != 0)
			throw new RuntimeError(status);

		// 指定が無ければデフォルトコンストラクタで生成する
		if (blob is null)
			blob = new T;

		// セグメントに分けて読み出す
		blob.resetWriteSegment();
		for (;;)
		{
			ushort written;
			ubyte[] segment = blob.getWriteSegment();
			if (segment is null)
				break;

			isc_get_segment(status.ptr, &handle, &written, cast(short)segment.length, segment.ptr);
			blob.appriseWrittenSize(written);
			if (status[1] != 0 && status[1] != isc_segment)
			{
				if (status[1] == isc_segstr_eof)
					break;
				else
					throw new RuntimeError(status);
			}
		}

		// blobをクローズ
		isc_close_blob(status.ptr, &handle);
		if (status[1] != 0)
			throw new RuntimeError(status);

		return blob;
	}

private:

	// executeしたステートメントハンドル
	Statement statement_;

	// fetchしたデータを格納するXSQLDA構造体
	XSQLDA* xsqlda_;

	// xsqldataに設定するデータ
	// ちょっとメモリは無駄になるけど安全のために8バイト境界に揃えておく
	ulong[] sqlDataBuffer_;

	// SQL indicatorの配列
	short[] sqlIndicatorBuffer_;

	/++
	 + コンストラクタ
	 +
	 + ユーザー任意の生成は禁止
	 +
	 + Params:
	 +  statementHandle = ステートメントハンドル
	 +  columnsNumber = 問い合わせ結果のカラム数
	 +
	 + Throws:
	 +  RuntimeError
	 +/
	this(Statement statement, short columnsNumber)
	{
		statement_ = statement;

		xsqlda_ = cast(XSQLDA*)(new ubyte[XSQLDA_LENGTH(columnsNumber)]).ptr;
		xsqlda_.ver = SQLDA_VERSION1;
		xsqlda_.sqln = columnsNumber;

		// 問い合わせ結果のカラム情報を取得する
		ISC_STATUS_ARRAY status;
		isc_dsql_describe(status.ptr, &statement_.handle_, 1, xsqlda_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		// 問い合わせ結果を格納する領域のサイズを計算する
		uint dataBufferLength;
		uint indicatorBufferLength;
		XSQLVAR* var = xsqlda_.sqlvar.ptr;
		for (short i; i<columnsNumber; ++i)
		{
			if ((var.sqltype & ~1) == SQL_VARYING)
				dataBufferLength += (var.sqllen + 2 + 7) >> 3;
			else
				dataBufferLength += (var.sqllen + 7) >> 3;

			if (var.sqltype & 1)
				++indicatorBufferLength;

			++var;
		}

		// sqldata用の領域を確保してポインタを設定
		sqlDataBuffer_ = new ulong[dataBufferLength];
		dataBufferLength = 0;
		var = xsqlda_.sqlvar.ptr;
		for (short i; i<columnsNumber; ++i)
		{
			var.sqldata = cast(ubyte*)(&sqlDataBuffer_[dataBufferLength]);

			if ((var.sqltype & ~1) == SQL_VARYING)
				dataBufferLength += (var.sqllen + 2 + 7) >> 3;
			else
				dataBufferLength += (var.sqllen + 7) >> 3;

			++var;
		}

		if (indicatorBufferLength == 0)
			return;

		// sqlind用の領域を確保してポインタを設定
		sqlIndicatorBuffer_ = new short[indicatorBufferLength];
		var = xsqlda_.sqlvar.ptr;
		for (short i, index; i<columnsNumber; ++i )
		{
			if ((var.sqltype & 1) != 0)
				var.sqlind = &sqlIndicatorBuffer_[index++];

			++var;
		}
	}
}

/++
 + ステートメント
 +
 + 使い方
 +  1. 実行するSQLと、使用する接続・トランザクションを指定し作成する
 +  2. prepare
 +  3. setメソッドでプレースホルダに値をセット
 +  4. execute. 結果が返るSQLならばResultSetを返す
 +  5. 必要ならば4で取得したResultSetをfetch
 +  6. finish
 +
 + 2 ～ 6の一連の操作は、接続・トランザクションが
 + 共に開いてある必要があります
 +/
class Statement
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  sql = 実行するSQL
	 +  transaction = トランザクション
	 +/
	this(string sql, Transaction transaction)
	in
	{
		assert(sql.length < 0x10000);
	}
	body
	{
		sql_ = sql;
		transaction_ = transaction;
	}

	/++
	 + SQL実行の準備をする
	 +/
	void prepare()
	{
		ISC_STATUS_ARRAY status;

		// ステートメント領域を確保する
		isc_dsql_allocate_statement(status.ptr, &transaction_.connection_.handle_, &handle_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		// 問い合わせ用パラメータ、結果領域が確保されているか確認する
		if (xsqlda_ !is null)
		{
			// 結果領域は既に確保されているのでprepareのみ行う
			XSQLDA* xsqlda = result_ is null ? null : result_.xsqlda_;
			isc_dsql_prepare(status.ptr, &transaction_.handle_, &handle_, cast(ushort)sql_.length, sql_.ptr, 1, xsqlda);
			if (status[0] == 1 && status[1] != 0)
				throw new RuntimeError(status);
		}
		else
		{
			// 問い合わせ・結果の領域を確保

			// まずは関数からの結果取得用領域を確保
			XSQLDA* outData = cast(XSQLDA*)(new byte[XSQLDA_LENGTH(1)]);
			outData.ver = SQLDA_VERSION1;
			outData.sqln = 0;
	
			// prepareして結果のデータ数を取得
			isc_dsql_prepare(status.ptr, &transaction_.handle_, &handle_, cast(ushort)sql_.length, sql_.ptr, 1, outData);
			if (status[0] == 1 && status[1] != 0)
				throw new RuntimeError(status);

			// 結果のデータ数を渡し、受け取る領域を生成
			if (outData.sqld > 0)
				result_ = new ResultSet(this, outData.sqld);


			// 問い合わせパラメータの数を取得
			XSQLDA* inData = cast(XSQLDA*)(new byte[XSQLDA_LENGTH(1)]);
			inData.ver = SQLDA_VERSION1;
			inData.sqln = 0;

			isc_dsql_describe_bind(status.ptr, &handle_, 1, inData);
			if (status[0] == 1 && status[1] != 0)
				throw new RuntimeError(status);

			// 問い合わせパラメータの数が足りなければ再確保する
			if (inData.sqld > inData.sqln)
			{
				xsqlda_ = cast(XSQLDA*)(new byte[XSQLDA_LENGTH(inData.sqld)]);
				xsqlda_.ver = SQLDA_VERSION1;
				xsqlda_.sqln = inData.sqld;

				isc_dsql_describe_bind(status.ptr, &handle_, 1, xsqlda_);
				if (status[0] == 1 && status[1] != 0)
					throw new RuntimeError(status);
			}
			else
				xsqlda_ = inData;

			if (xsqlda_.sqln > 0)
			{
				// 問い合わせ結果を格納する領域のサイズを計算する
				uint dataBufferLength;
				uint indicatorBufferLength;
				XSQLVAR* var = xsqlda_.sqlvar.ptr;
				for (short i; i<xsqlda_.sqln; ++i)
				{
					if ((var.sqltype & ~1) == SQL_VARYING)
						dataBufferLength += (var.sqllen + 2 + 7) >> 3;
					else
						dataBufferLength += (var.sqllen + 7) >> 3;

					if (var.sqltype & 1)
						++indicatorBufferLength;

					++var;
				}

				// sqldata用の領域を確保してポインタを設定
				sqlDataBuffer_ = new ulong[dataBufferLength];
				dataBufferLength = 0;
				var = xsqlda_.sqlvar.ptr;
				for (short i; i<xsqlda_.sqln; ++i)
				{
					var.sqldata = cast(ubyte*)(&sqlDataBuffer_[dataBufferLength]);

					if ((var.sqltype & ~1) == SQL_VARYING)
						dataBufferLength += (var.sqllen + 2 + 7) >> 3;
					else
						dataBufferLength += (var.sqllen + 7) >> 3;

					++var;
				}

				if (indicatorBufferLength == 0)
					return;

				// sqlind用の領域を確保してポインタを設定
				sqlIndicatorBuffer_ = new short[indicatorBufferLength];
				var = xsqlda_.sqlvar.ptr;
				for (short i, index; i<xsqlda_.sqln; ++i )
				{
					if ((var.sqltype & 1) != 0)
						var.sqlind = &sqlIndicatorBuffer_[index++];

					++var;
				}
			}
		}
	}

	/++
	 + プレースホルダに明示的にNULLをセットする
	 +
	 + stringとかblobとか、引数にnullを指定できる物はそのままnullを指定するだけでいいが、
	 + intとかnullを指定出来ないものはこのメソッドで明示的にNULLを指定する
	 + もちろんstring等もこれで明示的に指定してもよい
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +/
	void setNull(uint index)
	in
	{
		assert(index < xsqlda_.sqln );
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & 1) != 0);
		assert((xsqlda_.sqlvar.ptr + index).sqlind !is null);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;
		*(var.sqlind) = -1;
	}

	/++
	 + プレースホルダに文字列をセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = 文字列
	 +/
	void set(uint index, const(char)[] data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_TEXT || ((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_VARYING);
	}
	body
	{
		if (data is null)
			setNull(index);
		else
		{
			XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

			short length = cast(short)data.length;
			const(char)* read = data.ptr;

			ubyte* write;
			if ((var.sqltype & ~1) == SQL_VARYING)
			{
				*(cast(short*)var.sqldata) = length;

				write = var.sqldata + 2;
			}
			else
				write = var.sqldata;

			for (short i; i<length; ++i)
				*(write++) = *(read++);
		}
	}

	/++
	 + プレースホルダにshortをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, short data)
	in
	{
		assert(index < xsqlda_.sqln );
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_SHORT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(short*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにushortをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, ushort data)
	in
	{
		assert(index < xsqlda_.sqln );
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_SHORT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(ushort*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにintをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, int data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_LONG);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(int*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにuintをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, uint data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_LONG);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(uint*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにlongをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, long data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_INT64);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(long*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにulongをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, ulong data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_INT64);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(ulong*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにfloatをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, float data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_FLOAT);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(float*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにdoubleをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, double data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_DOUBLE);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		*(cast(double*)var.sqldata) = data;
	}

	/++
	 + プレースホルダにSysTimeをセットする
	 +
	 + Firebirdのtimestamp型は1万分の1秒までしか保持できません
	 + それより詳細なデータは切り捨てられます
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  data = セットするデータ
	 +/
	void set(uint index, SysTime data)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_TIMESTAMP);
	}
	body
	{
		XSQLVAR* var = xsqlda_.sqlvar.ptr + index;

		ISC_TIMESTAMP t;
		t.timestamp_date = cast(int)(data.stdTime / (24L * 60L * 60L * 10000000L)) - 678575;
		t.timestamp_time = cast(int)((data.stdTime % (24L * 60L * 60L * 10000000L)) / 1000L);
		*(cast(ISC_TIMESTAMP*)var.sqldata) = t;
	}

	/++
	 + プレースホルダにblobデータをセットする
	 +
	 + Params:
	 +  index = セットする対象のプレースホルダのインデックス(0基準)
	 +  blob = セットするデータ
	 +/
	void set(uint index, Blob blob)
	in
	{
		assert(index < xsqlda_.sqln);
		assert(((xsqlda_.sqlvar.ptr + index).sqltype & ~1) == SQL_BLOB);
	}
	body
	{
		if (blob is null)
			setNull(index);
		else
		{
			// blobデータのIDを書き込む位置を取得
			XSQLVAR* var = xsqlda_.sqlvar.ptr + index;
			ISC_QUAD* blobId = cast(ISC_QUAD*)var.sqldata;

			isc_blob_handle blobHandle = 0;
			ISC_STATUS_ARRAY status;

			// blobの作成
			isc_create_blob2(status.ptr, &transaction_.connection_.handle_, &transaction_.handle_, &blobHandle, blobId, 0, null);
			if (status[1] != 0)
				throw new RuntimeError(status);

			// セグメントが無くなるまで書き込み
			blob.resetReadSegment();
			for (const(ubyte)[] segment = blob.getReadSegment(); segment !is null; segment = blob.getReadSegment())
			{
				isc_put_segment(status.ptr, &blobHandle, cast(ushort)segment.length, segment.ptr);
				if (status[1] != 0)
				{
					// 一応クローズを試みる
					isc_close_blob(status.ptr, &blobHandle);
					throw new RuntimeError(status);
				}
			}

			// blobをクローズ
			isc_close_blob(status.ptr, &blobHandle);
			if (status[1] != 0)
				throw new RuntimeError(status);
		}
	}

	/++
	 + SQLの実行
	 +
	 + Returns:
	 +  問い合わせ結果取得インターフェース 結果の無いSQLの場合null
	 +
	 + Throws:
	 +  RuntimeError
	 +/
	ResultSet execute()
	{
		ISC_STATUS_ARRAY status;
		isc_dsql_execute(status.ptr, &transaction_.handle_, &handle_, 1, xsqlda_);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		return result_;
	}

	/++
	 + SQL実行の後始末をする
	 +/
	void finish()
	{
		ISC_STATUS_ARRAY status;
		isc_dsql_free_statement(status.ptr, &handle_, DSQL_drop);
		if (status[0] == 1 && status[1] != 0)
			throw new RuntimeError(status);

		handle_ = 0;
	}

private:

	// ステートメントハンドル
	isc_stmt_handle handle_;

	// SQL
	string sql_;

	// トランザクションオブジェクト
	Transaction transaction_;

	// 問い合わせ結果取得用インターフェース
	ResultSet result_;

	// 入力パラメータを格納するXSQLDA構造体
	XSQLDA* xsqlda_;

	// xsqldataに設定するデータ
	// ちょっとメモリは無駄になるけど安全のために8バイト境界に揃えておく
	ulong[] sqlDataBuffer_;

	// nullフラグ格納領域
	short[] sqlIndicatorBuffer_;
}

/++
 + blobデータ
 +/
interface Blob
{
	/++
	 + セグメントの開始位置を先頭に設定
	 +
	 + データベースに書き込むデータを読み出すセグメントの開始位置を先頭に設定する
	 +/
	void resetReadSegment();

	/++
	 + セグメントの開始位置を先頭に設定
	 +
	 + データベースから読み出したデータを書き込むセグメントの開始位置を先頭に設定する
	 +/
	void resetWriteSegment();
	
	/++
	 + セグメントを取得
	 +
	 + データベースに書き込むデータを読み出すセグメントを取得する
	 + 呼び出す毎に次のセグメントへ進む
	 +
	 + Returns:
	 +  セグメント 領域がもう無い場合は長さ0の配列ではなくnullを返す
	 +/
	const(ubyte)[] getReadSegment();

	/++
	 + セグメントを取得
	 +
	 + データベースから読み出したデータを書き込むセグメントを取得する
	 + こちらは自動的に次のセグメントへ進む機能は無い
	 + appriseWrittenSizeを使用して実際に書き込んだサイズを通知すること
	 +
	 + Returns:
	 +  セグメント 領域がもう無い場合は長さ0の配列ではなくnullを返す
	 +/
	ubyte[] getWriteSegment();

	/++
	 + セグメントに実際に書き込まれたサイズを知らせる
	 +/
	void appriseWrittenSize(ushort size);
}
