/++
 + テンプレートエンジン スイーツ(笑) Version: 0.4.1
 +
 + 基本機能モジュール
 +
 + Copyright(C) 2009-2011 Takuya Kurosawa
 +/
module jp.ku6.sweets.core;

import std.string;
import std.conv;


/++
 + Sweetsの出力を受け取るdelegate
 +
 + Sweetsのは処理結果として文字列を出力しますが、
 + その文字列を受け取るdelegateです
 +
 + メモリ領域への書き込み、ファイルへの書き出し、標準出力への出力などを実装します
 +
 + 出力された文字列はstringではなくconst(char)[]なので
 + 必ずコピーするようにしてください
 +/
alias void delegate(const(char)[]) SweetsPrinter;


/++
 + テンプレート中の処理対象要素を表すクラス
 +
 + SweetsParserが元となるテキストを解析すると
 + このSweetsElementのツリーが出力されます
 +
 + SweetsElementは"s:id"が付いたタグで囲まれた要素を表します
 +
 + それ以外の、通常テキストや"s:id"が付いていないタグは
 + ただの文字列として扱われ、出現直後のSweetsElementクラスに格納されています
 + (内部的な処理の都合なので、使用者が意識する必要はありません)
 +/
immutable
class SweetsElement
{
	/// 前のs:id付きタグ終了からs:id付きタグ出現までのテキスト(s:id無しタグも含む)
	string text;

	/// 要素に付けられたID
	string id;

	/// 要素名
	string name;

	/// 要素の属性
	string[string] attributes;

	/// 子要素の配列
	SweetsElement[] children;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  text = タグ出現までのテキスト
	 +  id = 要素のID
	 +  name = 要素名
	 +  attributes = 要素の属性
	 +  children = 子要素
	 +/
	this(string text, string name, string id, immutable(string[string]) attributes, immutable(SweetsElement[]) children)
	{
		this.text = text;
		this.id = id;
		this.name = name;
		this.attributes = attributes;
		this.children = children;
	}

	/++
	 + 開始タグを出力
	 +
	 + Params:
	 +  printer = 出力に使用するdelegate
	 +  close = trueを指定するとタグを"/>"で閉じる
	 +/
	void printStartTag(SweetsPrinter printer, bool close)
	{
		printStartTagWithOtherAttributes(printer, attributes, close);
	}

	/++
	 + 属性値を追加して出力
	 +
	 + インスタンスが保持している属性とキーが重複した場合引数の値で上書きされる
	 +
	 + Params:
	 +  printer = 出力に使用するdelegate
	 +  additionalAttributes = 追加の属性値
	 +  close = trueを指定するとタグを"/>"で閉じる
	 +/
	void printStartTagWithAdditionalAttributes(SweetsPrinter printer, const(string[string]) additionalAttributes, bool close)
	{
		if (name is null)
			return;

		string[string] margedAttributes;
		foreach (key, value; attributes)
			margedAttributes[key] = value;
		foreach (key, value; additionalAttributes)
			margedAttributes[key] = value;

		printStartTagWithOtherAttributes(printer, margedAttributes, close);
	}

	/++
	 + 属性を指定して開始タグを出力
	 +
	 + インスタンスが保持している属性は使わない
	 +
	 + Params:
	 +  printer = 出力に使用するdelegate
	 +  otherAttribute = 属性値
	 +  close = trueを指定するとタグを"/>"で閉じる
	 +/
	void printStartTagWithOtherAttributes(SweetsPrinter printer, const(string[string]) otherAttributes, bool close)
	{
		if (name is null)
			return;

		printer("<");
		printer(name);

		foreach (key, value; otherAttributes)
		{
			if (key.length < 2 || key[0] != 's' || key[1] != ':')
			{
				printer(" ");
				printer(key);
				printer("=\"");
				printer(value);
				printer("\"");
			}
		}

		if (close)
			printer("/>");
		else
			printer(">");
	}

	/++
	 + 終了タグを出力
	 +
	 + Params:
	 +  printer = 出力に使用するdelegate
	 +/
	const
	void printEndTag(SweetsPrinter printer)
	{
		printer("</");
		printer(name);	
		printer(">");
	}
}


/++
 + Sweetsに使用するテンプレート
 +
 + 元となるテキストを解析するとSweetsElementが複数出力されます
 + その複数のSweetsElementをひとまとめにしたクラスです
 +/
immutable
class SweetsTemplate
{
	/// テンプレート内のrootに位置する要素の配列
	SweetsElement[] elements;

	this(SweetsElement[] elements)
	{
		this.elements = elements;
	}
}


/++
 + 要素を処理するdelegate
 +
 + s:idが付いたタグで囲まれた要素(SweetsElement)を処理します
 + s:idとこのdelegateの紐付けはSweets.assignで行います
 +
 + 引数のSweetsPrinterに出力した内容が、s:idが付いたタグの代わりに出力されます
 +
 + 引数のSweetsElementが処理対象のSweetsElementです
 + 出力時に参照したい場合に使用します
 +
 + 引数のconst(SweetsNode[])は引数のSweetsElementの子ノードです
 + SweetsElementではなくSweetsNodeとなっているのは、子ノードもSweetsProcessorを持っている可能性があるためです
 + 主にs:id付きのタグをネストさせた場合に参照します
 +/
alias void delegate(SweetsPrinter, SweetsElement, const(SweetsNode[])) SweetsProcessor;


/++
 + テンプレート処理ノード
 +
 + SweetsElementとSweetsProcessorをペアにしたクラス
 + また子ノードもSweetsNodeとして保持している
 +/
class SweetsNode
{
	/// 要素に対して割り当てられた処理
	SweetsProcessor processor;

	/// 要素本体
	SweetsElement element;

	/// 子ノード
	SweetsNode[] children;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  e = 要素
	 +/
	this(SweetsElement e)
	{
		element = e;

		if (e.children.length > 0)
		{
			children = new SweetsNode[e.children.length];
			for (uint i; i < children.length; ++i)
				children[i] = new SweetsNode(e.children[i]);
		}
	}

	/++
	 + 指定されたSweetsPrinterに内容を出力
	 +
	 + Returns:
	 +  printer = 出力に使用するSweetsPrinter
	 +/
	const
	void print(SweetsPrinter printer)
	{
		if (element.text !is null)
			printer(element.text);

		// ID付きの要素かチェック
		if (element.id !is null)
		{
			// 処理が有ったら処理を行い、なければそのまま出力
			if (processor is null)
			{
				if (children.length > 0)
				{
					element.printStartTag(printer, false);

					foreach (child ; children)
						child.print(printer);

					element.printEndTag(printer);
				}
				else
					element.printStartTag(printer, true);
			}
			else
				processor(printer, element, children);
		}
	}

	/++
	 + ノードに設定された処理を実行し、結果をSweetsElementとして取得する
	 +
	 + 処理がセットされたs:idを処理し、無い状態にする
	 + 制約として、親が処理された場合子も処理された状態となる
	 +
	 + Returns:
	 +  SweetsElement
	 +/
	SweetsElement preprocess()
	{
		if (processor is null)
		{
			// 子が居れば処理する
			SweetsElement[] preprocessedChildren;
			if (children.length > 0)
			{
				string currentChild;
				for (uint i = 0; i < children.length; ++i)
				{
					SweetsElement child = children[i].preprocess();
					if (child.name is null)
					{
						if (currentChild is null)
							currentChild = child.text;
						else
							currentChild ~= child.text;
					}
					else
					{
						if (currentChild !is null)
						{
							preprocessedChildren ~= new SweetsElement(currentChild, null, null, null, null);
							currentChild = null;
						}

						preprocessedChildren ~= child;
					}
				}

				if (currentChild !is null)
					preprocessedChildren ~= new SweetsElement(currentChild, null, null, null, null);
			}

			return new SweetsElement(element.text, element.name, element.id, element.attributes, preprocessedChildren);
		}
		else
		{
			string buffer = element.text;
	
			void preprocessPrinter(const(char)[] data)
			{
				buffer ~= data.idup;
			}

			return new SweetsElement(buffer, null, null, null, null);
		}
	}

private:

	/+
	 + デフォルトコンストラクタ
	 +/
	this()
	{
		element = null;
	}
}


/++
 + HTMLテンプレートエンジン
 +
 + 使い方
 +
 + 1. HTMLテンプレートの作成
 +    任意のHTML文章を作成し、
 +    プログラム上から操作したいタグの属性のs:idにIDを指定する
 +    例: ID「hoge」を指定 → <div s:id="hoge"/>
 +
 + 2. SweetsParserオブジェクトを生成
 +    parse関数に、1.で用意したHTMLの文字列を渡し、SweetsTemplateオブジェクトを取得
 +  ----------------
 +  SweetsParser parser = new SweetsParser;
 +  SweetsTemplate tpl = parser.parse("<div s:id=\"hoge\">hogehoge</div>");
 +  ----------------
 +
 + 3. Sweetsオブジェクトを生成
 +    コンストラクタに、2.で取得したSweetsTemplateオブジェクトを渡す
 +  ----------------
 +  Sweets sweets = new Sweets(tpl);
 +  ----------------
 +
 + 4. 要素操作delegateを作成
 +    操作対象の要素と、表示用のdelegateを引数に取る関数(delegateに出来るもの)を用意する
 +  ----------------
 +  void processor(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
 +  {
 +  	element.printStartTag(printer, false);
 +  	printer("Hello work!!"); 
 +  	element.printEndTag(printer);
 +  }
 +  ----------------
 +
 + 5. 作成したdelegateをテンプレートのIDに割り付ける
 +  ----------------
 +  sweets.assign("hoge", &processor);
 +  ----------------
 +  
 + 6. Printerを作成
 +    表示する文字列を引数に取る関数(delegateに出来るもの)を用意する
 +  ----------------
 +  void printer(const(char)[] data)
 +  {
 +  	write(data);
 +  }
 +  ----------------
 +
 + 7. 表示
 +    要素に割り付けられた処理を行い、結果をPrinterに出力します
 +  ----------------
 +  sweets.print(&printer);
 +  ----------------
 +/
class Sweets
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  tpl = テンプレートデータ
	 +/
	this(SweetsTemplate tpl)
	{
		children_ = new SweetsNode[tpl.elements.length];
		for (uint i; i < children_.length; ++i)
			children_[i] = new SweetsNode(tpl.elements[i]);

		foreach (child; children_)
			reindexElements(child);
	}

	/++
	 + IDを指定して要素に処理を割り当てる
	 +
	 + 指定したIDのが存在しなければ何もしない
	 +
	 + Params:
	 +  id = ID
	 +  process = 処理
	 +/
	void assign(string id, SweetsProcessor processor)
	{
		SweetsNode[]* nodes = id in nodeIndex_;
		if (nodes !is null)
		{
			foreach (node; *nodes)
				node.processor = processor;
		}
	}

	/++
	 + 指定されたSweetsPrinterに内容を出力する
	 +
	 + Params:
	 +  printer = 出力先
	 +/
	void print(SweetsPrinter printer)
	{
		foreach (child; children_)
			child.print(printer);
	}

	/++
	 + SweetsProcessorがセットされたs:idタグのみを処理する
	 +
	 + Returns:
	 +  結果をSweetsTemplateとして返す
	 +/
	SweetsTemplate preprocess()
	{
		SweetsElement[] preprocessedChildren;
		if (children_.length > 0)
		{
			string currentChild;
			for (uint i = 0; i < children_.length; ++i)
			{
				SweetsElement child = children_[i].preprocess();
				if (child.name is null)
				{
					if (currentChild is null)
						currentChild = child.text;
					else
						currentChild ~= child.text;
				}
				else
				{
					if (currentChild !is null)
					{
						preprocessedChildren ~= new SweetsElement(currentChild, null, null, null, null);
						currentChild = null;
					}

					preprocessedChildren ~= child;
				}
			}

			if (currentChild !is null)
				preprocessedChildren ~= new SweetsElement(currentChild, null, null, null, null);
		}

		return new SweetsTemplate(preprocessedChildren);
	}

protected:

	// IDからノードを取得するインデックス
	SweetsNode[][string] nodeIndex_;

	// 子ノードの配列
	SweetsNode[] children_;

	/+
	 + デフォルトコンストラクタ
	 +/
	this()
	{
	}

	/+
	 + IDからノードを取得するインデックスを再生成する
	 +
	 + Params:
	 +  element = チェックする要素
	 +/
	void reindexElements(SweetsNode node)
	{
		if (node.element.id !is null)
		{
			SweetsNode[]* nodes = node.element.id in nodeIndex_;
			if (nodes !is null)
				*nodes ~= node;
			else
				nodeIndex_[node.element.id] = [node];
		}

		foreach (child; node.children)
			reindexElements(child);
	}
}
unittest
{
	char[] buffer;
	void printer(const(char)[] data)
	{
		buffer ~= data;
	}

	void processor(SweetsPrinter printer, SweetsElement element, const(SweetsNode[]) children)
	{
		element.printStartTag(printer, false);
		printer("Hello work!!"); 
		element.printEndTag(printer);
	}

	auto parser = new SweetsParser;
	auto tpl = parser.parse("<div s:id=\"hoge\">hogehoge</div><div s:id=\"fuga\">fugafuga</div>");
	auto sweets = new Sweets(tpl);

	sweets.assign("hoge", &processor);
	sweets.assign("fuga", &processor);

	buffer = [];
	sweets.print(&printer);
	assert(std.string.cmp("<div>Hello work!!</div><div>Hello work!!</div>", buffer.idup) == 0);
}


/++
 + テンプレートを解析するときのエラー
 +/
class SweetsParseError : Error
{
	/++
	 + エラー理由
	 +/
	enum Reason
	{
		START_TAG_NOT_CLOSED,           /// タグが開始されたが閉じられていない
		START_TAG_ATTRIBUTE_NOT_CLOSED, /// タグの属性の値がダブルクォートで始まったが、閉じるダブルクォートが見つからない
		START_TAG_HAS_ILLEGAL_SLASH,    /// タグの中に不正な/が見つかった
		END_TAG_NOT_CLOSED,             /// 閉じタグが開始されたが閉じられていない
		END_TAG_NOT_FOUND,              /// 開始タグがあるが閉じタグが見つからない
	}

	// エラー理由
	immutable
	Reason reason;

	// タグ名
	immutable
	string tag;

	// s:id値
	immutable
	string id;

	// 属性名
	immutable
	string attribute;

	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  reason = エラー理由
	 +  tag = タグ名
	 +  id = s:id値
	 +  attribute = 属性名
	 +/
	this(Reason reason, string tag, string id, string attribute)
	{
		super("Template parse error.");
		this.reason = reason;
		this.tag = tag;
		this.id = id;
		this.attribute = attribute;
	}
}


/++
 + Sweets用のテンプレート文字列を解析し、テンプレートデータを作成する
 +/
class SweetsParser
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		nodeStack_ = new TemporaryNode[4];
	}

	/++
	 + 解析処理
	 +
	 + 結果はnodesとelementIndexから取得する
	 +
	 + Params:
	 +  data = 解析するHTMLテンプレートデータ
	 +
	 + Throws:
	 +  SweetsParseError = 解析時にエラーが発生した場合
	 +/
	SweetsTemplate parse(string data)
	{
		// スタック初期化
		current_ = 0;

		// ルートノードをスタックに入れる
		nodeStack_[current_] = new TemporaryNode;

		// データの終了点を設定
		const(char*) end = data.ptr + data.length;

		// 通常テキストデータの開始点を設定
		textStart_ = data.ptr;

		const(char)* iterator;
		for (iterator = data.ptr; iterator < end; ++iterator)
		{
			if (*iterator == '<')
			{
				if (*(iterator+1) == '/')
					iterator = onEndTag(iterator, end);
				else
					iterator = onStartTag(iterator, end);
			}
		}

		// 残ったテキストを処理
		if (textStart_ < iterator)
		{
			auto lastNode = new TemporaryNode;
			lastNode.text = copyString(textStart_, iterator);

			nodeStack_[current_].children ~= lastNode;
		}

		// 全てのデータ処理後にスタックにタグが残っていたら閉じタグが無い状態なのでエラー
		if (current_ > 0)
		{
			TemporaryNode n = nodeStack_[current_];
			throw new SweetsParseError(SweetsParseError.Reason.END_TAG_NOT_FOUND, n.name, n.id, null);
		}

		// データを作成	
		return new SweetsTemplate(nodeStack_[0].toElement().children);
	}

private:

	/+
	 + 作業中のノードクラス
	 +/
	class TemporaryNode
	{
		// タグ出現前までのテキスト
		string text;

		// 要素に付けられたID
		string id;

		// 要素名
		string name;

		// 要素の属性
		string[string] attributes;

		// 子要素の配列
		TemporaryNode[] children;

		// タグを閉じるまでに同名タグが開いたカウント
		uint tagOpenCount;

		/+
		 + ノードオブジェクトに変換
		 +
		 + Returns:
		 +  ノードオブジェクト
		 +/
		immutable(SweetsElement) toElement()
		{
			SweetsElement[] cldren;
			for (uint i = 0; i < children.length; ++i)
				cldren ~= children[i].toElement();

			return new SweetsElement(text, name, id, cast(immutable(string[string]))attributes, cldren);
		}
	}

	// IDを指定する際の属性名
	static immutable
	string ID = "s:id";

	// テキストデータの開始点
	const(char)* textStart_;

	// 処理中のスタックの位置
	uint current_;

	// 作業中のノードのスタック
	TemporaryNode[] nodeStack_;

	/+
	 + スタックにノードを追加
	 +
	 + Params:
	 +  node = 追加するノード
	 +/
	void pushNode(TemporaryNode node)
	{
		// 足りなくなったらスタックをリサイズ
		if (++current_ >= nodeStack_.length)
			nodeStack_.length = nodeStack_.length * 2 + 1;

		nodeStack_[current_] = node;
	}

	/+
	 + スタックからノードを一つ削除
	 +
	 + 削除するノードは返さない
	 +/
	void popNode()
	in
	{
		assert(current_ > 0);
	}
	body
	{
		// 親ノードに子ノードを足す
		TemporaryNode parent = nodeStack_[current_-1];
		parent.children ~= nodeStack_[current_];

		-- current_;
	}

	/+
	 + 開始タグの処理
	 +
	 + Param:
	 +  iterator = 開始タグ開始文字のポインタ
	 +/
	const(char)* onStartTag(const(char)* iterator, const(char*) end)
	in
	{
		assert(*iterator == '<');
	}
	out
	{
		assert(*iterator == '>');
	}
	body
	{
		// タグ開始位置を記憶
		const(char*) tagStart = iterator;

		// 開始タグで閉じているかのフラグ
		bool closed = false;

		// タグ名称を取得
		for (++iterator; ; ++iterator)
		{
			if (iterator >= end)
				throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, copyString(tagStart + 1, iterator), null, null);
			else if (*iterator == '/')
			{
				if (*(iterator+1) != '>')
					throw new SweetsParseError(SweetsParseError.Reason.START_TAG_HAS_ILLEGAL_SLASH, copyString(tagStart + 1, iterator), null, null);
				break;
			}
			else if (*iterator == '>' || charactorIsSpace(*iterator))
				break;
		}
		string name = copyString(tagStart + 1, iterator);

		// タグ属性を取得
		string[string] attributes;
		while (true)
		{
			if (iterator >= end)
				throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, name, null, null);
			else if (*iterator == '/')
			{
				++iterator;
				if (*iterator != '>')
					throw new SweetsParseError(SweetsParseError.Reason.START_TAG_HAS_ILLEGAL_SLASH, name, null, null);
				closed = true;
				break;
			}
			else if (*iterator == '>')
			{
				break;
			}
			else if (!charactorIsSpace(*iterator))
			{
				const(char)* attributeNameStart = iterator;
				while (true)
				{
					if (iterator >= end)
						throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, name, null, null);

					if (*iterator == '>' || *iterator == '=' || charactorIsSpace(*iterator))
						break;

					++iterator;
				}

				string attributeName = copyString(attributeNameStart, iterator);
				if (*iterator == '>')
				{
					attributes[attributeName] = null;
					break;
				}

				for (; charactorIsSpace(*iterator); ++iterator)
				{
					if (iterator >= end)
						throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, name, null, null);
				}

				if (*iterator == '>')
				{
					attributes[attributeName] = null;
					break;
				}
	
				if (*iterator != '=')
				{
					attributes[attributeName] = null;
					continue;
				}

				for (++iterator; charactorIsSpace(*iterator); ++iterator)
				{
					if (iterator >= end)
						throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, name, null, null);
				}

				if (*iterator == '"')
				{
					const(char)* attributeValueStart = ++iterator;
					for (; *iterator != '"'; ++iterator)
					{
						if (iterator >= end)
							throw new SweetsParseError(SweetsParseError.Reason.START_TAG_ATTRIBUTE_NOT_CLOSED, name, null, attributeName);
					}
					attributes[attributeName] = copyString(attributeValueStart, iterator++);
				}
				else
				{
					const(char)* attributeValueStart = iterator;
					for (; !charactorIsSpace(*iterator); ++iterator)
					{
						if (iterator >= end)
							throw new SweetsParseError(SweetsParseError.Reason.START_TAG_NOT_CLOSED, name, null, null);
					}
					attributes[attributeName] = copyString(attributeValueStart, iterator);
				}
			}
			else
				++iterator;
		}

		if ((ID in attributes) is null)
		{
			// ノードタグでなくて、カレントと同名のタグならばカウントをインクリメント
			if (!closed && std.string.cmp(name, nodeStack_[current_].name) == 0)
				++nodeStack_[current_].tagOpenCount;
		}
		else
		{
			// ノードを作成
			auto node = new TemporaryNode;
			node.text = copyString(textStart_, tagStart);
			node.id = attributes[ID];
			attributes.remove(ID);
			node.name = name;
			node.attributes = attributes;

			pushNode(node);

			// 開始タグで閉じて居る場合は即POPする
			if (closed)
				popNode();

			textStart_ = iterator + 1;
		}

		return iterator;
	}

	/+
	 + 終了タグの処理
	 +
	 + Param:
	 +  iterator = 終了タグ開始文字のポインタ
	 +/
	const(char)* onEndTag(const(char)* iterator, const(char*) end)
	in
	{
		assert(*iterator == '<');
		assert(*(iterator+1) == '/');
	}
	out
	{
		assert(*iterator == '>');
	}
	body
	{
		const(char)* tagStart = iterator;
	
		bool close = true;

		// ノードのタグと同名か確認する
		const(char)[] name = nodeStack_[current_].name;
		if (name is null)
		{
			close = false;
		}
		else
		{
			for (uint i = 0; i < name.length; ++i)
			{
				if (name[i] != iterator[2+i])
				{
					close = false;
					break;
				}
			}

			char c = iterator[2+name.length];
			if (c != '>' && !charactorIsSpace(c))
				close = false;
		}

		// 閉じタグの終端まで移動
		while (true)
		{
			if (iterator >= end)
				throw new SweetsParseError(SweetsParseError.Reason.END_TAG_NOT_CLOSED, copyString(tagStart + 2, iterator), null, null);
			else if (*iterator == '>')
			{
				if (!close)
					return iterator;	// ノードのタグと同名でなければ終わり
				else
					break;
			}
			++iterator;
		}

		// 同名タグのスタックが空ならばノードの閉じタグと判定
		if (nodeStack_[current_].tagOpenCount == 0)
		{
			// 未処理のテキストがあるか判定
			if (textStart_ < tagStart)
			{
				// タグを持たないテキストのみのノードを作成
				auto lastNode = new TemporaryNode;
				lastNode.text = copyString(textStart_, tagStart);

				// 親ノードの最後の子として追加
				nodeStack_[current_].children ~= lastNode;
			}

			popNode();

			// テキスト開始位置をタグの直後にリセット
			textStart_ = iterator + 1;
		}
		else
			--nodeStack_[current_].tagOpenCount;	// 同名タグのスタックを減らす

		return iterator;
	}

	/+
	 + 文字が空白か判定
	 +
	 + Params:
	 +  c = 文字
	 +
	 + Returns:
	 +  空白文字の場合にtrueを返す
	 +/
	static bool charactorIsSpace(char c)
	{
		return c == ' ' || c == '\t' || c == '\r' || c == '\n';
	}

	/+
	 + 2点のポインタ間を文字列として取得する
	 +
	 + 返される文字列に開始点は含まれ、終了点は含まれない
	 + 開始点と終了点が等しい場合は空白文字列が返される
	 +
	 + Params:
	 +  start = 開始点
	 +  end = 終了点
	 +
	 + Returns:
	 +  2点間の文字列
	 +/
	static string copyString(const(char)* start, const(char)* end)
	in
	{
		assert(start <= end);
	}
	body
	{
		if (start == end)
			return "";

		size_t length = end - start;
		char[] buffer = new char[length];

		for (size_t i = 0; i < length; ++i)
			buffer[i] = start[i];

		return cast(immutable)buffer;
	}
}
