module jp.ku6.bbs.logic.ThreadLogic;

import std.datetime;
import jp.ku6.firebird.core;
import jp.ku6.firebird.extension;
import jp.ku6.bbs.model.Thread;
import jp.ku6.bbs.entity.ThreadEntity;
import jp.ku6.bbs.entity.ResponseEntity;
import jp.ku6.bbs.dao.ThreadDao;
import jp.ku6.bbs.dao.ResponseDao;

/++
 + スレッド管理クラス
 +/
class ThreadLogic
{
	/++
	 + コンストラクタ
	 +
	 + Params:
	 +  connectionPool = コネクションプール
	 +/
	this(ConnectionPool connectionPool)
	{
		connectionPool_ = connectionPool;
		transaction_ = new Transaction(null);
		threadDao_ = new ThreadDao(transaction_);
		responseDao_ = new ResponseDao(transaction_);
	}

	/++
	 + スレッドが存在するか確認
	 +
	 + Params:
	 +  id = スレッドID
	 +
	 + Returns:
	 +  スレッドが存在する場合にtrueを返す
	 +/
	bool threadExists(uint id)
	{
		if (id == 0)
			return false;

		mixin(SIMPLE_TRANSACTION_CONTROL);

		return threadDao_.countThreadById(id) == 1;
	}

	/++
	 + スレッド一覧情報を取得
	 +
	 + Returns:
	 +  スレッド一覧情報
	 +/
	Thread[] selectForListView()
	{
		mixin(SIMPLE_TRANSACTION_CONTROL);

		// データの取得
		return threadDao_.selectForListView();
	}

	/++
	 + IDを指定して1件取得
	 +
	 + Params:
	 +  id = ID
	 +
	 + Returns:
	 +  該当IDのレコード 無ければnullを返す
	 +/
	ThreadEntity selectById(uint id)
	{
		mixin(SIMPLE_TRANSACTION_CONTROL);
		return threadDao_.selectById(id);
	}

	/++
	 + スレッドIDを指定してレスを全件取得
	 +
	 + Params:
	 +  id = スレッドID
	 +
	 + Returns:
	 +  該当のレス
	 +/
	ResponseEntity[] selectResponsesById(uint id)
	{
		mixin(SIMPLE_TRANSACTION_CONTROL);
		return responseDao_.selectByThreadIdOrderByRegistered(id);
	}

	/++
	 + スレッド新規追加
	 +
	 + Params:
	 +  title = スレッドタイトル
	 +  name = >>1の名前
	 +  content = >>1の内容
	 +/
	void buildThread(string title, string name, string content)
	{
		mixin(SIMPLE_TRANSACTION_CONTROL);

		// 登録日時を取得
		auto now = Clock.currTime();

		// スレッド登録
		auto thread = new ThreadEntity;
		thread.id = threadDao_.selectNextSequence();
		thread.title = title;
		thread.registered = now;
		threadDao_.insert(thread);

		// >>1を登録
		auto response = new ResponseEntity;
		response.id = responseDao_.selectNextSequence();
		response.threadId = thread.id;
		response.name = name;
		response.content = content;
		response.registered = now;
		responseDao_.insert(response);
	}

	/++
	 + レスポンス書き込み
	 +
	 + Params:
	 +  threadId = 書き込み対象のスレッドID
	 +  name = 名前
	 +  content = 内容
	 +/
	void writeResponse(uint threadId, string name, string content)
	{
		mixin(SIMPLE_TRANSACTION_CONTROL);

		// 登録日時を取得
		auto now = Clock.currTime();

		// >>1を登録
		auto response = new ResponseEntity;
		response.id = responseDao_.selectNextSequence();
		response.threadId = threadId;
		response.name = name;
		response.content = content;
		response.registered = now;
		responseDao_.insert(response);
	}


protected:

	// コネクションプール
	ConnectionPool connectionPool_;

	// トランザクション
	Transaction transaction_;

	// スレッドDAO
	ThreadDao threadDao_;

	// レスポンスDAO
	ResponseDao responseDao_;

	immutable static
	string SIMPLE_TRANSACTION_CONTROL = `
		// コネクションの取得と返却処理
		transaction_.connection = connectionPool_.lend(3);
		scope(exit)
		{
			connectionPool_.collect(transaction_.connection);
			transaction_.connection = null;
		}

		// トランザクションの開始と終了処理
		transaction_.begin();
		scope(failure) transaction_.rollback();
		scope(success) transaction_.commit();
	`;
}
